########### R function: SolveThreeLevelSparseLeastSquares ###########

# For solving a three-level sparse matrix least squares problem
# according to Algorithm 4 of Nolan, Menictas & Wand (2018).

# Last changed: 17 SEP 2018

SolveThreeLevelSparseLeastSquares <- function(b,B,Bdot,Bdotdot)
{
   # Extract dimension variables:

   m <- length(b)
   nVec <- as.numeric(unlist(lapply(b,length)))
   p <- ncol(B[[1]][[1]]) ; q1 <- ncol(Bdot[[1]][[1]])
   q2 <- ncol(Bdotdot[[1]][[1]])

   # Initialise 'omega7' and 'Omega8' matrices:

   omega7 <- NULL ;  Omega8 <- NULL

   # Perform the QR decomposition sweep over both indices:

   c1 <- vector("list",m)  ;  C1 <- vector("list",m)
   d1 <- vector("list",m)  ;  D1 <- vector("list",m) ; D1dot <- vector("list",m)
   RL1 <- vector("list",m) ;  RL2 <- vector("list",m)

   for (i in 1:m)
   {

      d1[[i]] <- vector("list",nVec[i]) ; D1[[i]] <- vector("list",nVec[i])
      D1dot[[i]] <- vector("list",nVec[i]) ; RL2[[i]] <- vector("list",nVec[i])

      # Initialise 'omega9', 'Omega10' and 'Omega11' matrices:

      omega9 <- NULL  ;  Omega10 <- NULL  ;  Omega11 <- NULL

      for (j in 1:nVec[i])
      {
          # Obtain the QR decomposition of Bdotdot_ij and obtain the
          # R_ij matrices:

          qrObj <- qr(Bdotdot[[i]][[j]])   ;   RL2[[i]][[j]] <- qr.R(qrObj)

          # Form the d, D, Ddot and D1dot matrices and update 'omega9', 'Omega10'
          # and 'Omega11':

          d0ij <- qr.qty(qrObj,b[[i]][[j]])
          D0ij <- qr.qty(qrObj,B[[i]][[j]])
          Ddot0ij <- qr.qty(qrObj,Bdot[[i]][[j]])

          d1[[i]][[j]] <- d0ij[1:q2] ; d2ij <- d0ij[-(1:q2)]
          omega9 <- c(omega9,d2ij)

          D1[[i]][[j]] <- D0ij[1:q2,] ; D2ij <- D0ij[-(1:q2),]
          Omega10 <- rbind(Omega10,D2ij)

          D1dot[[i]][[j]] <- Ddot0ij[1:q2,] ; Ddot2ij <- Ddot0ij[-(1:q2),]
          Omega11 <- rbind(Omega11,Ddot2ij)
      }

      # Obtain the QR decomposition of 'Omega11' and obtain the R_i matrices:

      qrObj <- qr(Omega11)   ;   RL1[[i]] <- qr.R(qrObj)

      # Form the c and C matrices and update 'omega7' and 'Omega8':

      c0i <- qr.qty(qrObj,omega9)
      C0i <- qr.qty(qrObj,Omega10)

      c1[[i]] <- c0i[1:q1] ; c2i <- c0i[-(1:q1)]
      omega7 <- c(omega7,c2i)

      C1[[i]] <- C0i[1:q1,] ; C2i <- C0i[-(1:q1),]
      Omega8 <- rbind(Omega8,C2i)
   }

   # Obtain solutions for the (1,1) block:

   qrObj <- qr(Omega8)   ;   R <- qr.R(qrObj)

   cvec <- qr.qty(qrObj,omega7)[1:p]

   x1 <- backsolve(R,cvec) ;  Asup11 <- backsolve(R,t(backsolve(R,diag(p))))

   # Perform the final solution sweep over both indices:

   x2L1 <- vector("list",m)  ;  Asup12L1 <- vector("list",m)  ; Asup22L1 <- vector("list",m)
   x2L2 <- vector("list",m)  ;  Asup12L2 <- vector("list",m)  ; Asup1comma2L2 <- vector("list",m)
   Asup22L2 <- vector("list",m)

   for (i in 1:m)
   {
      x2L2[[i]] <- vector("list",nVec[i])  ;  Asup12L2[[i]] <- vector("list",nVec[i])

      Asup1comma2L2[[i]] <- vector("list",nVec[i]) ;  Asup22L2[[i]] <- vector("list",nVec[i])

      x2L1[[i]] <- backsolve(RL1[[i]],(c1[[i]] - crossprod(t(C1[[i]]),x1)))

      Asup12L1[[i]] <- - tcrossprod(Asup11,backsolve(RL1[[i]],C1[[i]]))

      Asup22L1[[i]] <- backsolve(RL1[[i]],(t(backsolve(RL1[[i]],diag(q1)))
                                           - crossprod(t(C1[[i]]),Asup12L1[[i]])))

      for (j in 1:nVec[i])
      {
          x2L2[[i]][[j]] <- backsolve(RL2[[i]][[j]],(d1[[i]][[j]] - crossprod(t(D1[[i]][[j]]),x1)
                                                     - crossprod(t(D1dot[[i]][[j]]),x2L1[[i]])))

          Asup12L2[[i]][[j]] <- -t(backsolve(RL2[[i]][[j]],(crossprod(t(D1[[i]][[j]]),Asup11) +
                                                           tcrossprod(D1dot[[i]][[j]],Asup12L1[[i]]))))

          Asup1comma2L2[[i]][[j]] <- -t(backsolve(RL2[[i]][[j]],(crossprod(t(D1[[i]][[j]]),Asup12L1[[i]]) +
                                                           tcrossprod(D1dot[[i]][[j]],Asup22L1[[i]]))))

          Asup22L2[[i]][[j]] <- backsolve(RL2[[i]][[j]],t(backsolve(RL2[[i]][[j]],diag(q2)))
                                           - crossprod(t(D1[[i]][[j]]),Asup12L2[[i]][[j]])
                                           - crossprod(t(D1dot[[i]][[j]]),Asup1comma2L2[[i]][[j]]))
      }
   }

   # Obtain the log|A| value:
   
   logDetA <- 2*sum(log(abs(diag(R))))
   for (i in 1:m) 
   {
      logDetA <- logDetA + 2*sum(log(abs(diag(RL1[[i]]))))
      for (j in 1:nVec[i])
         logDetA <- logDetA + 2*sum(log(abs(diag(RL2[[i]][[j]]))))
   }

   # Return solutions:

   return(list(x1 = x1,Asup11 = Asup11,x2L1 = x2L1,Asup22L1 = Asup22L1,Asup12L1 = Asup12L1,
               x2L2 = x2L2,Asup22L2 = Asup22L2,Asup12L2 = Asup12L2,Asup1comma2L2 = Asup1comma2L2,
               logDetA = logDetA))
}

########### End of SolveThreeLevelSparseLeastSquares ############
