########### R function: threeLevelMFVB ###########

# For performing mean field variational Bayesian inference
# for three-level mixed models. This corresponds to Algorithm 6
# of  Nolan, Menictas & Wand (2018).

# Last changed: 18 JUN 2020

threeLevelMFVB <- function(yList, XList, ZL1List, ZL2List, mu.beta, Sigma.beta,
                           s.sigsq, nu.sigsq = 1, nu.SigmaL1 = 2, nu.SigmaL2 = 2,
                           s.SigmaL1, s.SigmaL2, tolerance = 1e-10, maxIter = 1000,
                           quiet = FALSE)
{
   # Obtain dimension variables:

   m <- length(yList)
   p <- ncol(XList[[1]][[1]])
   q1 <- ncol(ZL1List[[1]][[1]])
   q2 <- ncol(ZL2List[[1]][[1]])

   nVec <- as.numeric(unlist(lapply(yList,length)))
   oMat <- matrix(length(yList[[1]][[1]]),m,nVec[1])

   # Obtain b,B and Bdot lists:

   b <- vector("list", m)
   B <- vector("list", m)
   Bdot <- vector("list", m)
   Bdotdot <- vector("list", m)

   for (i in 1:m)
   {
      b[[i]] <- vector("list", nVec[i])
      B[[i]] <- vector("list", nVec[i])
      Bdot[[i]] <- vector("list", nVec[i])
      Bdotdot[[i]] <- vector("list", nVec[i])
   }

   # Do MFVB iterations:

   mu.q.recip.sigsq <- 1
   mu.q.recip.a.sigsq <- 1
   M.q.inv.SigmaL1 <- diag(q1)
   M.q.inv.SigmaL2 <- diag(q2)
   M.q.inv.A.SigmaL1 <- diag(q1)
   M.q.inv.A.SigmaL2 <- diag(q2)

   xi.q.sigsq <- nu.sigsq + sum(oMat)
   xi.q.SigmaL1 <- nu.SigmaL1 + (2 * q1) - 2 + m
   xi.q.SigmaL2 <- nu.SigmaL2 + (2 * q2) - 2 + sum(nVec)
   xi.q.a.sigsq <- nu.sigsq + 1
   xi.q.A.SigmaL1 <- nu.SigmaL1 + q1
   xi.q.A.SigmaL2 <- nu.SigmaL2 + q2

   Lambda.q.A.SigmaL1 <- matrix(0, q1, q1)
   Lambda.q.A.SigmaL2 <- matrix(0, q2, q2)

   itnum <- 0
   converged <- FALSE
   logMLcurr <- -Inf
   while (!converged)
   {
      itnum <- itnum + 1
      if (!quiet) cat("Performing iteration number ",itnum,".\n",sep = "")

      # Set current log(ML) value to previous one:

      logMLprev <- logMLcurr

      for (i in 1:m)
      {
         for (j in 1:nVec[i])
         {

            # Obtain b, B, Bdot and Bdotdot lists:

            b[[i]][[j]] <- c(yList[[i]][[j]] * sqrt(mu.q.recip.sigsq),
                              (1/sqrt(sum(nVec))) * matrixSqrt(solve(Sigma.beta)) %*% mu.beta,
                              rep(0, q1 + q2))
            B[[i]][[j]] <- rbind(XList[[i]][[j]] * sqrt(mu.q.recip.sigsq),
                                  (1/sqrt(sum(nVec))) * matrixSqrt(solve(Sigma.beta)),
                                  matrix(0, q1, p),
                                  matrix(0, q2, p))
            Bdot[[i]][[j]] <- rbind(ZL1List[[i]][[j]] * sqrt(mu.q.recip.sigsq),
                                     matrix(0, p, q1),
                                     matrixSqrt((1/nVec[i]) * M.q.inv.SigmaL1),
                                     matrix(0, q2, q1))
            Bdotdot[[i]][[j]] <- rbind(ZL2List[[i]][[j]] * sqrt(mu.q.recip.sigsq),
                                        matrix(0, p, q2),
                                        matrix(0, q1, q2),
                                        matrixSqrt(M.q.inv.SigmaL2))
         }
      }

      # Call the SolveThreeLevelSparseLeastSquares algorithm:

      S8 <- SolveThreeLevelSparseLeastSquares(b, B, Bdot, Bdotdot)

      mu.q.beta <- S8$x1   ;   Sigma.q.beta <- S8$Asup11

      mu.q.uL1 <- S8$x2L1  ;   Sigma.q.uL1 <- S8$Asup22L1
      E.q.betauL1 <- S8$Asup12L1

      mu.q.uL2 <- S8$x2L2  ;  Sigma.q.uL2 <- S8$Asup22L2
      E.q.betauL2 <- S8$Asup12L2
      E.q.uL1uL2 <- S8$Asup1comma2L2

      two_norm_terms <- 0
      uL1_tcrossprod_terms <- 0
      uL2_tcrossprod_terms <- 0

      for (i in 1:m)
      {
         uL1_tcrossprod_terms <- uL1_tcrossprod_terms + tcrossprod(mu.q.uL1[[i]]) + Sigma.q.uL1[[i]]

         for (j in 1:nVec[i])
         {
            residVec <- (yList[[i]][[j]] - as.vector(XList[[i]][[j]]%*%mu.q.beta)
                         - as.vector(ZL1List[[i]][[j]]%*%mu.q.uL1[[i]])
                         - as.vector(ZL2List[[i]][[j]]%*%mu.q.uL2[[i]][[j]]))

            two_norm_terms  <- ( two_norm_terms
                                 + sum(residVec^2)
                                 + sum(diag(tcrossprod(crossprod(XList[[i]][[j]]),
                                       Sigma.q.beta)))
                                 + sum(diag(tcrossprod(crossprod(ZL1List[[i]][[j]]),
                                       Sigma.q.uL1[[i]])))
                                 + 2*sum(diag(tcrossprod(crossprod(ZL1List[[i]][[j]],
                                         XList[[i]][[j]]), t(E.q.betauL1[[i]]))))
                                 + 2*sum(diag(tcrossprod(crossprod(ZL2List[[i]][[j]],
                                         XList[[i]][[j]]), t(E.q.betauL2[[i]][[j]]))))
                                 + 2*sum(diag(tcrossprod(crossprod(ZL1List[[i]][[j]],
                                         ZL2List[[i]][[j]]), t(E.q.uL1uL2[[i]][[j]])))))

            uL2_tcrossprod_terms <- (uL2_tcrossprod_terms + tcrossprod(mu.q.uL2[[i]][[j]])
                                     + Sigma.q.uL2[[i]][[j]])
         }
      }

      lambda.q.sigsq <- mu.q.recip.a.sigsq + two_norm_terms
      Lambda.q.SigmaL1 <- M.q.inv.A.SigmaL1 + uL1_tcrossprod_terms
      Lambda.q.SigmaL2 <- M.q.inv.A.SigmaL2 + uL2_tcrossprod_terms

      mu.q.recip.sigsq <- xi.q.sigsq/lambda.q.sigsq

      # Update M_{q(SigmaL1^{-1})} and M_{q(SigmaL2^{-1})}:

      M.q.inv.SigmaL1 <- (xi.q.SigmaL1 - q1 + 1) * solve(Lambda.q.SigmaL1)
      M.q.inv.SigmaL2 <- (xi.q.SigmaL2 - q2 + 1) * solve(Lambda.q.SigmaL2)

      # Update q*(a.sigsq) parameter:

      lambda.q.a.sigsq <- mu.q.recip.sigsq + (1/(nu.sigsq * s.sigsq^2))
      mu.q.recip.a.sigsq <- xi.q.a.sigsq / lambda.q.a.sigsq

      # Update q*(A_SigmaL1) and q*(A_SigmaL2) parameters:

      diag(Lambda.q.A.SigmaL1) <- diag(M.q.inv.SigmaL1) +  (1/(nu.SigmaL1 * s.SigmaL1^2))
      M.q.inv.A.SigmaL1 <- xi.q.A.SigmaL1 * solve(Lambda.q.A.SigmaL1)

      diag(Lambda.q.A.SigmaL2) <- diag(M.q.inv.SigmaL2) +  (1/(nu.SigmaL2 * s.SigmaL2^2))
      M.q.inv.A.SigmaL2 <- xi.q.A.SigmaL2 * solve(Lambda.q.A.SigmaL2)

      logMLcurr <- threeLevelMargLogLik(p, m, q1, q2, nVec, oMat, XList, L1List, L2List,
                                        two_norm_terms, uL1_tcrossprod_terms, uL2_tcrossprod_terms,
                                        nu.sigsq, s.sigsq, mu.q.recip.sigsq, xi.q.sigsq, lambda.q.sigsq,
                                        mu.q.recip.a.sigsq, xi.q.a.sigsq, lambda.q.a.sigsq, Sigma.beta,
                                        mu.beta, mu.q.beta, Sigma.q.beta,  M.q.inv.SigmaL1, xi.q.SigmaL1,
                                        Lambda.q.SigmaL1, nu.SigmaL1, s.SigmaL1, M.q.inv.A.SigmaL1,
                                        xi.q.A.SigmaL1, Lambda.q.A.SigmaL1, M.q.inv.SigmaL2, xi.q.SigmaL2,
                                        Lambda.q.SigmaL2, nu.SigmaL2, s.SigmaL2, M.q.inv.A.SigmaL2,
                                        xi.q.A.SigmaL2, Lambda.q.A.SigmaL2)

      # Compute relative error:

      relErr <- abs((logMLcurr/logMLprev) - 1)

      # Check 'converged' conditions:

      if ((relErr<tolerance)|(itnum>=maxIter)) converged <- TRUE
   }

   return(list(mu.q.beta = mu.q.beta, Sigma.q.beta = Sigma.q.beta, mu.q.uL1 = mu.q.uL1,
              Sigma.q.uL1 = Sigma.q.uL1, mu.q.uL2 = mu.q.uL2, Sigma.q.uL2 = Sigma.q.uL2,
              xi.q.sigsq = xi.q.sigsq, lambda.q.sigsq = lambda.q.sigsq,
              E.q.betauL1 = E.q.betauL1, E.q.betauL2 = E.q.betauL2, E.q.uL1uL2 = E.q.uL1uL2,
              xi.q.a.sigsq = xi.q.a.sigsq, lambda.q.a.sigsq = lambda.q.a.sigsq, xi.q.SigmaL1 = xi.q.SigmaL1,
              xi.q.SigmaL2 = xi.q.SigmaL2, Lambda.q.SigmaL1 = Lambda.q.SigmaL1, Lambda.q.SigmaL2 = Lambda.q.SigmaL2,
              xi.q.A.SigmaL1 = xi.q.A.SigmaL1, xi.q.A.SigmaL2 = xi.q.A.SigmaL2, Lambda.q.A.SigmaL1 = Lambda.q.A.SigmaL1,
              Lambda.q.A.SigmaL2 = Lambda.q.A.SigmaL2))
}

########### End of threeLevelMFVB ############
