########## R script: threeLevelMFVBexamples ##########

# For doing simulated data examples of streamlined mean field
# variational Bayes for linear mixed models with three-level
# random effects structure.

# Last changed: 15 JUN 2020

# Set tasks:

generateData <- TRUE
obtainMFVBfits <- TRUE
plotMFVBfits <- TRUE

# source required scripts:

source("matrixSqrt.r")
source("SolveThreeLevelSparseLeastSquares.r")
source("threeLevelMFVB.r")
source("threeLevelMargLogLik.r")

# Load required packages:

library(MASS) ; library(lattice)

# Set example number and seed:

exampNum <- 3
set.seed(1)

if (exampNum==1)
{
   m <-  6
   nVec <-  rep(8,m)
   oMat <-  matrix(25,m,nVec[1])
   layoutVec <- c(8,6)
   stripCexVal <- 0.6
}
if (exampNum==2)
{
   m <-  12
   nVec <-  rep(20,m)
   oMat <-  matrix(8,m,nVec[1])
   layoutVec <- c(20,12)
   stripCexVal <- 0.3
}
if (exampNum==3)
{
   m <-  20
   nVec <-  rep(50,m)
   oMat <-  matrix(30,m,nVec[1])
   layoutVec <- c(50,20)
   stripCexVal <- 0.3
}

if (!any(exampNum == 1:3))
   stop(call. = FALSE,"exampNum must be either 1, 2 or 3.\n")

numObs <- sum(oMat)

# Set true values of parameters:

beta0True <- 0.58
beta1True <- 1.89

#SigmaL1True <- matrix(c(0.32, 0.09, 0.09, 0.42), 2, 2)
#SigmaL2True <- matrix(c(0.30, 0.12, 0.12, 0.25), 2, 2)

SigmaL1True <- matrix(c(0.42, -0.09, -0.09, 0.52), 2, 2)
SigmaL2True <- matrix(c(0.80, -0.24, -0.24, 0.75), 2, 2)

sigsqEpsTrue <- 0.7

# Set dimension variables:

ncX <- 2
ncZL1 <- 2
ncZL2 <- 2

if (generateData)
{
   # Set hyperparameters:

   mu.beta <- rep(0,ncX)
   Sigma.beta <- 1e10*diag(ncX)
   s.sigsq <- 1e5
   nu.sigsq <- 1
   s.SigmaL1 <- rep(1e5,ncZL1)
   nu.SigmaL1 <- 2
   s.SigmaL2 <- rep(1e5,ncZL2)
   nu.SigmaL2 <- 2

   # Generate data:

   x <- NULL ; y <- NULL
   idnumL1 <- NULL; idnumL2 <- NULL
   currStt <- 1
   reBlockInds <- vector("list", length=m)
   yList <- vector("list", length=m)
   XList <- vector("list", length=m)
   ZL1List <- vector("list", length=m)
   ZL2List <- vector("list", length=m)
   uRL1 <- mvrnorm(m, rep(0,2), SigmaL1True)

   for (i in 1:m)
   {
      uRL2 <- mvrnorm(nVec[i],rep(0,2),SigmaL2True)

      yList[[i]] <- vector("list",length=nVec[i])
      XList[[i]] <- vector("list",length=nVec[i])
      ZL1List[[i]] <- vector("list",length=nVec[i])
      ZL2List[[i]] <- vector("list",length=nVec[i])

      for(j in 1:nVec[i])
      {

         idnumL1 <- c(idnumL1,rep(i,oMat[i,1]))
         idnumL2 <- c(idnumL2,rep(j,oMat[i,j]))

         xCurr <- runif(oMat[i,j])

         muCurr <- ((beta0True + uRL1[i,1] + uRL2[j,1]) +
                     (beta1True + uRL1[i,2] + uRL2[j,2])*xCurr)
         yCurr <- rnorm(oMat[i,j],muCurr,sqrt(sigsqEpsTrue))
         x <- c(x,xCurr)
         y <- c(y,yCurr)

         currEnd <- currStt + length(xCurr) - 1
         reBlockInds[[i]][j] <- list(currStt:currEnd)
         currStt <- currEnd + 1

         yList[[i]][[j]] <- yCurr
         XList[[i]][[j]] <- cbind(1,xCurr)
         ZL1List[[i]][[j]] <- cbind(1,xCurr)
         ZL2List[[i]][[j]] <- cbind(1,xCurr)
      }
   }
}

if (obtainMFVBfits)
{
   # Obtain streamlined mean field variational Bayes fit object:

   threeLevelMFVBObj <- threeLevelMFVB(
            yList = yList,
            XList = XList,
            ZL1List = ZL1List,
            ZL2List = ZL2List,
            mu.beta = mu.beta,
            Sigma.beta = Sigma.beta,
            s.sigsq = s.sigsq,
            nu.sigsq = nu.sigsq,
            s.SigmaL1 = s.SigmaL1,
            s.SigmaL2 = s.SigmaL2,
            nu.SigmaL1 = nu.SigmaL1,
            nu.SigmaL2 = nu.SigmaL2)

   # Extract q-density parameter vectors:

   mu.q.beta <- threeLevelMFVBObj$mu.q.beta
   Sigma.q.beta <- threeLevelMFVBObj$Sigma.q.beta
   mu.q.uL1 <- threeLevelMFVBObj$mu.q.uL1
   mu.q.uL2 <- threeLevelMFVBObj$mu.q.uL2
   Sigma.q.uL1 <- threeLevelMFVBObj$Sigma.q.uL1
   Sigma.q.uL2 <- threeLevelMFVBObj$Sigma.q.uL2
   E.q.betauL1 <- threeLevelMFVBObj$E.q.betauL1
   E.q.betauL2 <- threeLevelMFVBObj$E.q.betauL2
   E.q.uL1uL2 <- threeLevelMFVBObj$E.q.uL1uL2
}

if (plotMFVBfits)
{
   numOuters <- length(unique(idnumL1))
   idnumL1fac <- idnumL1
   for (i in 1:numOuters)
      idnumL1fac[idnumL1==i] <- paste("outer",as.character(i))
   idnumL1fac <- as.factor(idnumL1fac)
   idnumL1fac <- ordered(idnumL1fac,paste("outer",1:numOuters))

   numInners <- length(unique(idnumL2))
   idnumL2fac <- idnumL2
   for (i in 1:numInners)
      idnumL2fac[idnumL2==i] <- paste("inner",as.character(i))
   idnumL2fac <- as.factor(idnumL2fac)
   idnumL2fac <- ordered(idnumL2fac,paste("inner",1:numInners))

   fitCol <- "darkmagenta"

   xg <- seq(0,1,length=31)
   Xg <- cbind(1,xg)
   XgXgXg <- cbind(Xg,Xg,Xg)

   xgList <- vector("list",m)
   meangList <- vector("list",m)
   lowgList <- vector("list",m)
   uppgList <- vector("list",m)

   INDSbeta <- 1:ncX

   uL1Stt <- ncX + 1

   for (i in 1:m)
   {
      uL1End <- uL1Stt + ncZL1 - 1
      INDSuL1 <- uL1Stt:uL1End

      xgList[[i]] <- vector("list",nVec[i])
      meangList[[i]] <- vector("list",nVec[i])

      uL2Stt <- uL1End + 1

      for (j in 1:nVec[i])
      {
         uL2End <- uL2Stt + ncZL2 - 1
         INDSuL2 <- uL2Stt:uL2End

         xgList[[i]][[j]] <- xg
         meangList[[i]][[j]] <- ((mu.q.beta[1] + mu.q.uL1[[i]][1] + mu.q.uL2[[i]][[j]][1])
                                  + (mu.q.beta[2] + mu.q.uL1[[i]][2] + mu.q.uL2[[i]][[j]][2])*xg)

         covMatFull <- matrix(0,(ncX + ncZL1 + ncZL2),(ncX + ncZL1 + ncZL2))

         covMatFull[INDSbeta,INDSbeta] <- Sigma.q.beta
         covMatFull[INDSuL1,INDSuL1] <- Sigma.q.uL1[[i]]
         covMatFull[INDSbeta,INDSuL1] <- E.q.betauL1[[i]]
         covMatFull[INDSuL1,INDSbeta] <- t(E.q.betauL1[[i]])
         covMatFull[INDSuL2,INDSuL2] <- Sigma.q.uL2[[i]][[j]]
         covMatFull[INDSbeta,INDSuL2] <- E.q.betauL2[[i]][[j]]
         covMatFull[INDSuL2,INDSbeta] <- t(E.q.betauL2[[i]][[j]])
         covMatFull[INDSuL1,INDSuL2] <- E.q.uL1uL2[[i]][[j]]
         covMatFull[INDSuL2,INDSuL1] <- t(E.q.uL1uL2[[i]][[j]])

         sdCurr <-  sqrt(diag(XgXgXg%*%covMatFull%*%t(XgXgXg)))

         lowgList[[i]][[j]] <- meangList[[i]][[j]] - qnorm(0.975)*sdCurr
         uppgList[[i]][[j]] <- meangList[[i]][[j]] + qnorm(0.975)*sdCurr
      }
   }

   figFits <- xyplot(y~x|idnumL2fac*idnumL1fac,groups=idnumL2fac,
                       data=data.frame(x=x,y=y,idnumL1fac=idnumL1fac,idnumL2fac=idnumL2fac),
                       layout=layoutVec,main="",
                       strip=strip.custom(par.strip.text= list(cex = stripCexVal)),
                       par.settings = list(layout.heights = list(strip = stripCexVal)),
                       xlab="x",ylab="y",as.table=TRUE,
                       panel=function(x,y,subscripts,groups)
                       {
                          iPan <- panel.number()

                          i <- rep(1:numOuters,each=numInners)[iPan]
                          j <- rep(1:numInners,numOuters)[iPan]

                          panel.grid()

                          panel.superpose(x[order(x)],y[order(x)],subscripts,groups,
                                         type="b",col="dodgerblue",pch=1,cex=0.1)

                          panel.xyplot(xgList[[i]][[j]],meangList[[i]][[j]],
                                       col=fitCol,type="l",lwd=2)

                          panel.xyplot(xgList[[i]][[j]],lowgList[[i]][[j]],
                                       col=fitCol,type="l",lwd=2,lty=2)

                          panel.xyplot(xgList[[i]][[j]],uppgList[[i]][[j]],
                                       col=fitCol,type="l",lwd=2,lty=2)

                       })

   print(figFits)
}

######### End of threeLevelMFVBexamples ##########
