########### R function: twoLevelMFVB ###########

# For performing mean field variational Bayesian inference
# for two-level mixed models. This corresponds to Algorithm 6
# of  Nolan, Menictas & Wand (2018).

# Last changed: 18 JUN 2020

twoLevelMFVB <- function(yList, XList, ZList, mu.beta, Sigma.beta, s.sigsq, nu.sigsq = 1,
                         nu.Sigma = 2, s.Sigma, tolerance = 1e-10, maxIter = 1000,
                         quiet = FALSE)
{
   # Set dimension values:

   m <- length(yList)
   p <- ncol(XList[[1]])
   q <- ncol(ZList[[1]])

   nVec <- unlist(lapply(yList, length))

   # Obtain b,B and Btilde lists:

   b <- vector("list", m)
   B <- vector("list", m)
   Bdot <- vector("list", m)

   # Do MFVB iterations:

   mu.q.recip.sigsq <- 1
   mu.q.recip.a.sigsq <- 1
   M.q.inv.Sigma <- diag(q)
   M.q.inv.A.Sigma <- diag(q)

   xi.q.sigsq <- nu.sigsq + sum(nVec)
   xi.q.Sigma <- nu.Sigma + (2 * q) - 2 + m
   xi.q.a.sigsq <- nu.sigsq + 1
   xi.q.A.Sigma <- nu.Sigma + q

   Lambda.q.A.Sigma <- matrix(0, q, q)

   itnum <- 0
   converged <- FALSE
   logMLcurr <- -Inf
   while (!converged)
   {
      itnum <- itnum + 1
      if (!quiet) cat("Performing iteration number ",itnum,".\n",sep = "")

      # Set current log(ML) value to previous one:

      logMLprev <- logMLcurr

      for (i in 1:m)
      {
         # Obtain b,B and Bdot lists:

         b[[i]] <- c(yList[[i]] * sqrt(mu.q.recip.sigsq),
                     (matrixSqrt(solve(m * Sigma.beta)) %*% mu.beta),
                     rep(0,q))

         B[[i]] <- rbind(XList[[i]] * sqrt(mu.q.recip.sigsq),
                          matrixSqrt(solve(m * Sigma.beta)),
                          matrix(0,q,p))

         Bdot[[i]] <- rbind(ZList[[i]]*sqrt(mu.q.recip.sigsq),
                             matrix(0,p,q),
                             matrixSqrt(M.q.inv.Sigma))
      }

      # Call the SolveTwoLevelSparseLeastSquares algorithm:

      S2 <- SolveTwoLevelSparseLeastSquares(b,B,Bdot)

      mu.q.beta <- S2$x1   ;   Sigma.q.beta <- S2$Asup11
      mu.q.u <- S2$x2   ;   Sigma.q.u <- S2$Asup22
      E.q.betau <- S2$Asup12

      two_norm_terms <- 0
      u_tcrossprod_terms <- 0

      for (i in 1:m)
      {
         residVec <- (yList[[i]] - as.vector(XList[[i]] %*% mu.q.beta)
                                  - as.vector(ZList[[i]] %*% mu.q.u[[i]]))

         two_norm_terms <- ( two_norm_terms
                             + sum(residVec^2)
                             + sum(diag(tcrossprod(crossprod(XList[[i]]), Sigma.q.beta)))
                             + sum(diag(tcrossprod(crossprod(ZList[[i]]), Sigma.q.u[[i]])))
                             + 2*sum(diag(tcrossprod(crossprod(ZList[[i]],
                                     XList[[i]]),t(E.q.betau[[i]])))) )

         u_tcrossprod_terms <- ( u_tcrossprod_terms + tcrossprod(mu.q.u[[i]]) + Sigma.q.u[[i]] )
      }

      lambda.q.sigsq <- mu.q.recip.a.sigsq + two_norm_terms
      Lambda.q.Sigma <- M.q.inv.A.Sigma + u_tcrossprod_terms

      # Update mu_{q(1/sigsq)}:

      mu.q.recip.sigsq <- xi.q.sigsq/lambda.q.sigsq

      # Update M_{q(Sigma^{-1})}:

      M.q.inv.Sigma <- (xi.q.Sigma - q + 1) * solve(Lambda.q.Sigma)

      # Update q*(a.sigsq) parameter:

      lambda.q.a.sigsq <- mu.q.recip.sigsq + (1 / (nu.sigsq * (s.sigsq^2)))
      mu.q.recip.a.sigsq <- xi.q.a.sigsq / lambda.q.a.sigsq

      # Update q*(A_Sigma) parameters:

      Lambda.q.A.Sigma <- diag(diag(M.q.inv.Sigma)) + diag((1/(nu.Sigma * (s.Sigma^2))))
      M.q.inv.A.Sigma <- xi.q.A.Sigma * solve(Lambda.q.A.Sigma)

      # Obtain current log(ML):

      logMLcurr <- twoLevelMargLogLik(p, m, q, nVec, ZList, two_norm_terms,
         u_tcrossprod_terms, nu.sigsq, s.sigsq, mu.q.recip.sigsq, xi.q.sigsq,
         lambda.q.sigsq, mu.q.recip.a.sigsq, xi.q.a.sigsq, lambda.q.a.sigsq,
         Sigma.beta, mu.beta, mu.q.beta, Sigma.q.beta, M.q.inv.Sigma,
         xi.q.Sigma, Lambda.q.Sigma, nu.Sigma, s.Sigma, M.q.inv.A.Sigma,
         xi.q.A.Sigma, Lambda.q.A.Sigma)

      # Compute relative error:

      relErr <- abs((logMLcurr/logMLprev) - 1)

      # Check 'converged' conditions:

      if ((relErr<tolerance)|(itnum>=maxIter)) converged <- TRUE
   }

   return(list(mu.q.beta = mu.q.beta, Sigma.q.beta = Sigma.q.beta, mu.q.u = mu.q.u,
               Sigma.q.u = Sigma.q.u,E.q.betau = E.q.betau, xi.q.sigsq = xi.q.sigsq,
               lambda.q.sigsq = lambda.q.sigsq,xi.q.a.sigsq = xi.q.a.sigsq,
               lambda.q.a.sigsq = lambda.q.a.sigsq, xi.q.Sigma = xi.q.Sigma,
               Lambda.q.Sigma = Lambda.q.Sigma, xi.q.A.Sigma = xi.q.A.Sigma,
               Lambda.q.A.Sigma = Lambda.q.A.Sigma))
}

########### End of twoLevelMFVB ############
