########## R script: twoLevelMFVBexamples ##########

# For doing simulated data examples of streamlined mean field
# variational Bayes for linear mixed models with two-level
# random effects structure.

# Last changed: 18 JUN 2020

# Set tasks:

generateData <- TRUE
obtainMFVBfits <- TRUE
plotMFVBfits <- TRUE

# Load required functions:

source("matrixSqrt.r")
source("SolveTwoLevelSparseLeastSquares.r")
source("twoLevelMFVB.r")
source("twoLevelMargLogLik.r")

# Load required packages:

library(MASS) ; library(lattice)

# Set example number and seed:

exampNum <- 4
set.seed(1)

if (exampNum == 1)
{
   m <- 20
   nVec <- rep(15,m)
   layoutVec <- c(5,4)
}

if (exampNum == 2)
{
   m <- 102
   nVec <- NULL
   for (i in 1:m)
      nVec[i] <- sample(25:35,1,replace=TRUE)
   layoutVec <- c(17,6)
}

if (exampNum == 3)
{
   m <- 200
   nVec <- NULL
   for (i in 1:m)
      nVec[i] <- sample(150:200,1,replace=TRUE)
   layoutVec <- c(20,10)
}

if (exampNum == 4)
{
   m <- 1000
   nVec <- NULL
   for (i in 1:m)
      nVec[i] <- sample(50:60,1,replace=TRUE)
   layoutVec <- c(50,20)
}

if (!any(exampNum == 1:4))
   stop(call. = FALSE,"exampNum must be either 1, 2, 3 or 4.\n")

if (generateData)
{
   # Set hyperparameters:

   mu.beta <- rep(0,2)
   Sigma.beta <- 1e10*diag(2)
   s.sigsq <- 1e5
   nu.sigsq <- 1
   s.Sigma <- rep(1e5, 2)
   nu.Sigma <- 2

   # Set true values of parameters:

   beta0True <- 0.58
   beta1True <- 1.89
   SigmaTrue <- matrix(c(2.58,0.22,0.22,1.73),2,2)
   sigsqEpsTrue <- 0.3

   # Generate data:

   x <- NULL; y <- NULL
   idnum <- NULL; currStt <- 1
   reBlockInds <- vector("list", length=m)
   ZTy <- vector("list", length=m)

   u <- mvrnorm(m,rep(0,2),SigmaTrue)
   for (i in 1:m)
   {
      idnum <- c(idnum,rep(i,nVec[i]))

      xCurr <- runif(nVec[i])
      muCurr <- ((beta0True + u[i,1]) +
                 (beta1True + u[i,2])*xCurr)
      yCurr <- rnorm(nVec[i],muCurr,sqrt(sigsqEpsTrue))
      x <- c(x,xCurr);  y <- c(y,yCurr)

      currEnd <- currStt + length(xCurr) - 1
      reBlockInds[i] <- list(currStt:currEnd)
      currStt <- currEnd + 1

      ZTy[[i]] <- crossprod(cbind(rep(1,nVec[i]),xCurr),yCurr)
   }
   numObs <- sum(nVec)
   X <- cbind(1,x)
   Z <- X
   ncX <- ncol(X)
   ncZ <- ncol(Z)
   XTX <- crossprod(X)
   XTy <- crossprod(X,y)

   yList <- vector("list", length=m)
   XList <- vector("list", length=m)
   ZList <- vector("list", length=m)

   for (i in 1:m)
   {
      yList[[i]] <- y[reBlockInds[[i]]]
      XList[[i]] <- X[reBlockInds[[i]],]
      ZList[[i]] <- X[reBlockInds[[i]],]
   }
}

if (obtainMFVBfits)
{
   # Obtain streamlined mean field variational Bayes fit object:

   twoLevelMFVBObj <- twoLevelMFVB(yList = yList,
                                   XList = XList,
                                   ZList = ZList,
                                   mu.beta = mu.beta,
                                   Sigma.beta = Sigma.beta,
                                   s.sigsq = s.sigsq,
                                   nu.sigsq = nu.sigsq,
                                   nu.Sigma = nu.Sigma,
                                   s.Sigma = s.Sigma)

   # Extract q-density parameter vectors:

   mu.q.beta <- twoLevelMFVBObj$mu.q.beta
   Sigma.q.beta <- twoLevelMFVBObj$Sigma.q.beta
   mu.q.u <- twoLevelMFVBObj$mu.q.u
   Sigma.q.u <- twoLevelMFVBObj$Sigma.q.u
   E.q.betau <- twoLevelMFVBObj$E.q.betau
}

if (plotMFVBfits)
{
   fitCol <- c("darkmagenta","deeppink")

   xg <- seq(0,1,length=31)
   Xg <- cbind(1,xg)
   XgXg <- cbind(Xg,Xg)

   xgList <- vector("list",m)
   meangList <- vector("list",m) ; lowgList <- vector("list",m) ; uppgList <- vector("list",m)

   for (i in 1:m)
   {
      xgList[[i]] <- xg
      meangList[[i]] <- (mu.q.beta[1] + mu.q.u[[i]][1]) + (mu.q.beta[2] + mu.q.u[[i]][2])*xg

      INDSbeta <- 1:ncX ; INDSu <- (ncX+1):(ncX+ncZ)

      covMatFull <- matrix(0,(ncX+ncZ),(ncX+ncZ))
      covMatFull[INDSbeta,INDSbeta] <- Sigma.q.beta
      covMatFull[INDSu,INDSu] <- Sigma.q.u[[i]]
      covMatFull[INDSbeta,INDSu] <- E.q.betau[[i]]
      covMatFull[INDSu,INDSbeta] <- t(E.q.betau[[i]])

      sdCurr <- sqrt(diag(XgXg%*%covMatFull%*%t(XgXg)))

      lowgList[[i]] <- meangList[[i]] - qnorm(0.975)*sdCurr
      uppgList[[i]] <- meangList[[i]] + qnorm(0.975)*sdCurr
   }

   figFits <- xyplot(y~x|as.factor(idnum),groups=idnum,
                     data=data.frame(x=x,y=y,idnum=idnum),
                     layout=layoutVec,main="",strip=FALSE,
                     xlab="x",ylab="y",
                     panel=function(x,y,subscripts,groups)
                     {
                        i <- panel.number()
                        panel.grid()

                        panel.superpose(x[order(x)],y[order(x)],subscripts,groups,
                                        type="b",col="dodgerblue",pch=1,cex=0.1)

                        panel.xyplot(xgList[[i]],meangList[[i]],
                                     col=fitCol[1],type="l",lwd=2)
                        panel.xyplot(xgList[[i]],lowgList[[i]],
                                     col=fitCol[1],type="l",lwd=2,lty=2)
                        panel.xyplot(xgList[[i]],uppgList[[i]],
                                     col=fitCol[1],type="l",lwd=2,lty=2)
                    })

   print(figFits)
}

######### End of twoLevelMFVBexamples ##########
